(* print.sml
 _______________________________________________________________________
|									|
|			      Projet LCS				|
|			  Groupe OLC, CNRS/LAAS				|
|			 7, Av. du Colonel Roche			|
|			   31077 Toulouse Cedex				|
|_______________________________________________________________________|

Oppen style pretty printing functions.

*)




(* interface

structure Print =
   val print_init : unit -> unit
   val pretty_printing : bool -> unit
   val print_depth : int -> unit
   val print_width : int -> unit

   exception Print : exn
   val print_begin : bool * int -> unit
   val print_token : string -> unit
   val print_break : int * int -> unit
   val print_end : unit -> unit

   val print_newline : unit -> unit
   val print_line : string -> unit

   val print_unit : unit -> unit
   val print_string : string -> unit
   val print_bool : bool -> unit
   val print_int : int -> unit
   val print_mp : mp -> unit
   val print_real : real -> unit

   val print_list : bool * bool ->
		    string * string * string -> 
		    ('a -> 'b) -> 
		    'a list ->
		    unit
*)





(* implementation *)

needs "stack";
needs "queue";

structure Print =
struct


local

open Stack;
open Queue;

(* type token. --------------------------------------------------------	*)

datatype brkind =
	  consis			(* consistent			*)
	| incons			(* inconsistent			*)
	| fits;

datatype ptoken =			(* to-be-printed tokens		*)
	  str_ptok of string		(* string token			*)
	| brk_ptok of int * int		(* width * offset		*)
	| beg_ptok of int * brkind	(* indent * break-kind		*)
	| end_ptok;			(* end				*)


(* constants, parameters and variables. -------------------------------	*)

val PAGE_WIDTH = ref 72;		(* line width			*)
val LARGE = 999999; 			(* large enough			*)

val CURR_DEPTH = ref ~1			(* current depth of "begins"	*)
and MAX_DEPTH = ref 16			(* deepest block to print       *)
and PRETTYON = ref true;		(* pretty printing on/off	*)

val SPACE_LEFT = ref (!PAGE_WIDTH)	(* space left on this line	*)
and PRINTED_SO_FAR = ref 1		(* total width already printed	*)
and QUEUED_SO_FAR = ref 1;		(* total width ever put in queue *)


(* The print stack. ---------------------------------------------------	*)

exception Print_stack = Pop;
val (clear_print_stack,
     _,
     push_print_stack,
     pop_print_stack,
     top_print_stack) = newostack () : (brkind * int) ostack;


(* The print queue. ---------------------------------------------------	*)

exception Print_queue = Deq;
val (clear_queue,
     null_queue,
     enqueue,
     dequeue,
     head) = newoqueue () : (ptoken * int ref) oqueue;


(* The scan stack. ----------------------------------------------------	*)

local
val SSTACK = ref ([]: (ptoken * int ref) list)
fun last [x] = (x,[])
  | last (h::t) = let val (b,tt) = last t in (b,(h::tt)) end
in
exception Scan_stack
fun clear_scan_stack () = SSTACK := []
and null_scan_stack () =  null (!SSTACK)
and push_scan_stack x =   SSTACK := x:: !SSTACK
and pop_scan_stack () = case !SSTACK of
	(h::t) => (SSTACK := t; h) | _ => raise Scan_stack
and top_scan_stack () = case !SSTACK of
	(h::_) => h | _ => raise Scan_stack
and pop_bottom_scan_stack () = case !SSTACK of
	  (_::_) => let val (b,t) = last (!SSTACK) in SSTACK := t; b end
	| _ => raise Scan_stack
end;


(* print. -------------------------------------------------------------	*)

fun indent 0 = () | indent n = (write " "; indent (n-1));

fun break_new_line (brk_ptok (_,off)) = (* breaks line, indenting new line *)
 	 (SPACE_LEFT := (snd (top_print_stack ())) - off
	; write "\n"
	; indent ((!PAGE_WIDTH) - !SPACE_LEFT));

fun break_same_line (brk_ptok (w,_)) = (* breaks on current line *)
	(SPACE_LEFT := !SPACE_LEFT - w; indent w);

fun printtk (tk,sz) = (* prints token tk, of size sz *)
	case tk of
	  str_ptok s => (SPACE_LEFT := !SPACE_LEFT - sz; write s)
	| beg_ptok (i,k) =>
		let val ofst = !SPACE_LEFT - i
		    and brk = if sz > !SPACE_LEFT then k else fits
		in push_print_stack (brk,ofst)
		end
	| end_ptok => (pop_print_stack (); ())
	| brk_ptok (w,i) =>
		(case (fst (top_print_stack ())) of
		  consis => break_new_line tk
		| incons => (if sz > !SPACE_LEFT then break_new_line
			    else break_same_line) tk
		| fits => break_same_line tk);

fun advance () =
	let val go = ref true
	in while (!go) andalso not (null_queue ())
	   do let val (tk,ref sz) = head ()
	      in if sz >= 0
	         then (dequeue (); printtk(tk,sz)
		     ; PRINTED_SO_FAR := !PRINTED_SO_FAR + sz)
	         else go := false
	      end
	end;

fun enqueue_tok (tok,l) = 
	if null_scan_stack () then printtk (tok,l)
	else (enqueue (tok,ref l)
            ; QUEUED_SO_FAR := !QUEUED_SO_FAR + l
            ; while !QUEUED_SO_FAR - !PRINTED_SO_FAR > !SPACE_LEFT
	      do let val (_,szr) = pop_bottom_scan_stack ()
	         in szr := LARGE; advance ()
		 end);


in (* local *)

(* pretty printing primitives. ----------------------------------------	*)

fun print_token "" = ()  (* prints a litteral string *)
  | print_token s = 
	let val l = string_length s
	in if !PRETTYON
	   then enqueue_tok (str_ptok s,l)
	   else (if l < !SPACE_LEFT
	         then (SPACE_LEFT := !SPACE_LEFT -l; write s)
	         if l < !PAGE_WIDTH
	         then (write "\n"; write (chr 9); write s;
		       SPACE_LEFT := !PAGE_WIDTH - (8+l))
	         else (write s; write "\n"; write (chr 9); 
		       SPACE_LEFT := !PAGE_WIDTH - 8))
	end;

fun print_string s = print_token s
and print_int n = print_token (string_of_int n)
and print_mp n = print_token (string_of_mp n)
and print_real n = print_token (string_of_real n)
and print_unit u = print_token (string_of_unit u)
and print_bool b = print_token (string_of_bool b);

exception Print;
fun print_begin (b,i) = (* opens a new block, indents if necessary *)
	if !PRETTYON
	then (inc CURR_DEPTH;
	      if !CURR_DEPTH < !MAX_DEPTH
	      then (if null_scan_stack()
		    then (PRINTED_SO_FAR := 1; QUEUED_SO_FAR := 1);
	            let val k = (beg_ptok (i,if b then consis else incons),
				           ref (~(!QUEUED_SO_FAR)))
		    in enqueue k; push_scan_stack k
		    end)
	      else (enqueue_tok (str_ptok "&",1); raise Print))
	else (inc CURR_DEPTH;
	      if !CURR_DEPTH >= !MAX_DEPTH then raise Print);

fun print_end () = (* closes a block, setting sizes of its sub-blocks *)
	if !PRETTYON
	then (if !CURR_DEPTH < !MAX_DEPTH
	      then (if null_scan_stack () then printtk (end_ptok,0)
	            else enqueue (end_ptok,ref 0))
            ; if not (null_scan_stack())
              then let val (tk,szr) = pop_scan_stack ()
	           in (szr := !QUEUED_SO_FAR + !szr
	             ; if can (fn brk_ptok _ =>()) tk andalso
			  not (null_scan_stack ())
		       then let val (_,szr2) = pop_scan_stack ()
	        	    in szr2 := !QUEUED_SO_FAR + !szr2
			    end
             	     ; if null_scan_stack () then advance ())
 		   end
     	    ; dec CURR_DEPTH);

fun print_break (w,off) =
	if !PRETTYON
	then (if !CURR_DEPTH < !MAX_DEPTH
	      then (if null_scan_stack () then printtk(brk_ptok (w,off),w)
                    else (QUEUED_SO_FAR := !QUEUED_SO_FAR + w
	                ; let val (tk,szr) = top_scan_stack ()
	                  in case tk of
		           brk_ptok _ => (pop_scan_stack ();
					  szr := !QUEUED_SO_FAR + !szr)
		  	 | _ => ()
	         	  end)
                  ; let val k = (brk_ptok (w,off),ref (~(!QUEUED_SO_FAR)))
		    in enqueue k; push_scan_stack k
		    end))
	else (if w<>0 then write " ");

fun print_init () = (* initializes pretty printer *)
	 (clear_queue ()
	; clear_print_stack ()
	; clear_scan_stack ()
	; CURR_DEPTH := ~1
	; SPACE_LEFT := !PAGE_WIDTH
	; PRINTED_SO_FAR := 1
	; QUEUED_SO_FAR := 1);


fun print_newline () = (* new line after printing all queued text *)
	(* beware: dont use it within blocks; it reinits then printer *)
	 (QUEUED_SO_FAR := LARGE
	; advance ()
	; write "\n"
	; print_init ());

fun pretty_printing b = (* turns pretty-printer on/off *)
	(PRETTYON := b; print_init ());

fun print_depth n = MAX_DEPTH := n
and print_width n = PAGE_WIDTH := n;


(* derived printers. --------------------------------------------------	*)

fun print_line string = (* beware: should not be used within blocks *)
	(print_string string; print_newline ());

local
fun iter_print seps off p l n =
  let fun go [] _ = ()
        | go _ 0 = (print_string seps; print_break (0,off); print_string "...")
        | go (h::t) n = (print_string seps; print_break (0,off); p h; go t (n-1))
  in go l n
  end;
in
fun print_list (brk,cut) (begs,seps,ends) p l =
	let val off = size begs in
	 (print_begin (brk,0)
	; print_string begs
	; case l of [] => ()
	     | (h::t) => (p h
		; if cut
		  then iter_print seps off p t (!MAX_DEPTH - (!CURR_DEPTH + 1))
		  else (map (fn x => (print_string seps
			            ; print_break (0,off); p x)) t; ()))
	; print_string ends
	; print_end ())
	end;
end;

end; (* local *)
end; (* structure *)




