(* term.sml
 _______________________________________________________________________
|									|
|			      Projet LCS				|
|			  Groupe OLC, CNRS/LAAS				|
|			 7, Av. du Colonel Roche			|
|			   31077 Toulouse Cedex				|
|_______________________________________________________________________|

notes:

* capabilities of terminals [preliminary version]

*)




(* signature:

structure Term =
  exception Terminfo : string -> exn
  val beep_win : window -> unit -> unit
  val delchar_win : window -> unit -> unit
  val home_win : window -> unit -> unit
  val clear_win : window -> unit -> unit
  val rvideo_win : window -> unit -> unit
  val move_win : window -> int * int -> unit

  ... to be completed ...

*)






structure Term = struct

exception Terminfo of string;

local
(* reading and analyzing terminfo compiled files *)
(* see term.doc for the codes and meaning of various flags, 
   numbers and strings encoding terminal capabilities. *)
(* the analysis functions below fail with Terminfo <message>
   if the particular capability searched for is not provided. *)
(* variable ti is a terminfo compiled file *)

fun short s n = ord (substring(s,n,1)) + 256*(ord(substring (s,n+1,1)));
fun terminfo_flag ti n = (* nth boolean of ti (n >= 0) *)
	ord (substring (ti,12+short ti 2+n,1));
fun terminfo_number ti n = (* nth number of ti *)
        let val d = 12+short ti 2+short ti 4
            val r = short ti (d+d mod 2+2*n)
        in if r < 0
	   then raise Terminfo ("number "^makestring n^"undefined")
           else r
        end;
fun terminfo_string ti n = (* offset in ti of nth string of ti *)
        let val d = 12 + short ti 2 + short ti 4
            val s = d + d mod 2 + 2*short ti 6
            val i = short ti (s + 2*n)
        in if i <= 0 then raise Terminfo "string sequence unavailable"
           else s + 2*short ti 8 + i
        end;
 
(* terminfo string decoding *)
(* terminal manipulation interface *)
(* *** This implementation is experimental ***
   only a few capabilities implemented below. Please extend according
   to your needs. In addition:
   - outputf below is fine for the *_win functions implemented, but
     would not properly handle any terminfo string. 
   - outputf should NOT use makestring, but should use instead a version 
     of it that return C-style string representations of constants.
     [e.g. should return "-1" for ~1, rather than "~1". Also, outputf 
     assumes that its string arg ends with \000. 
*)
datatype constant = I of int | R of real (* | ... etc, *);
overload const unconst;
fun const i = I i;
fun const r = R r;
fun unconst (I i) = i | unconst _ = raise Terminfo "can't unpack";
fun unconst (R r) = r | unconst _ = raise Terminfo "can't unpack";
nonoverload const unconst;
fun send (oo,ti,offset,args) =
	let val i = ref 0
	    val a = ref 0
	    val c = ref (substring (ti,offset,1))
	    val off = ref offset
	    fun adv () = (inc off; c := substring (ti,!off,1))
	in while !c <> "\000"
	   do if !c = "%"
	      then (case (adv(); !c) of
                  "i" => (i := 1; adv())
              	| "d" => (puts (oo, makestring (!i
                               + unconst (Vector.sub (args,!a)):int), 0);
                        adv(); inc a)
	        | "f" => (puts (oo, makestring (real (!i)
                               + unconst (Vector.sub (args,!a)):real), 0);
                        adv(); inc a)
  		| "p" => (adv(); adv())
		| _ => raise Terminfo "unable to decode terminfo string, sorry")
	      else (puts (oo,!c,0); adv())
	end;

fun get_terminfo t = (* gets terminfo dict for TERM t *)
	let val i = substring (t,0,1)
	    val ii = open_in ("/usr/share/lib/terminfo/"^i^"/"^t)
		     handle Io _ =>
		     open_in ("/usr/lib/terminfo/"^i^"/"^t)
		     handle Io _ =>
		     open_in ("/usr/share/terminfo/"^i^"/"^t)
		     handle Io _ => raise Terminfo ("can't find terminfo file for "^t)
	    val ti = input(ii,4096)
	in close_in ii; ti
	end;
in

(* user interface *)

(* to avoid loading terminfo at each call, all these functions are curried,
   the first application returns a closure that captures the terminfo dict.
   The first arg should be applied only once.
   typical use:
	fun f w ... =
		let val beep = beep_win w
		    val rvideo = rvideo_win w
		    val ...
		in ... beep () ... rvideo () ... beep () ... etc
		end;
*)

fun beep_win w = (* beep, visual or sound *)
	let val ti = get_terminfo (win_term w)
	    val bw = terminfo_string ti 1
	in fn () => send (win_out w, ti, bw, #[])
	end;
fun delchar_win w = (* delete character *)
	let val ti = get_terminfo (win_term w)
	    val dc = terminfo_string ti 21
	in fn () => send (win_out w, ti, dc, #[])
	end;
fun clear_win w = (* clear screen *)
	let val ti = get_terminfo (win_term w)
	    val cl = terminfo_string ti 5
	in fn () => send (win_out w, ti, cl, #[])
	end;
fun home_win w = (* move cursor to home *)
	let val ti = get_terminfo (win_term w)
	    val ho = terminfo_string ti 12
	in fn () => send (win_out w, ti, ho, #[])
	end;
fun rvideo_win w = (* reverse video on *)
	let val ti = get_terminfo (win_term w)
	    val rv = terminfo_string ti 34
	in fn () => send (win_out w, ti, rv, #[])
	end;
fun move_win w = (* move cursor to position (i,j) *)
        let val ti = get_terminfo (win_term w)
	    val mw = terminfo_string ti 10
	in fn (i:int,j:int) => send (win_out w, ti, mw, #[const i, const j])
	end;
(* MORE TO BE PROVIDED *)
end;

end;
